import {ComponentType} from "@angular/cdk/portal";
import {
  ChangeDetectorRef,
  OnDestroy,
  OnInit,
  ViewChild,
} from "@angular/core";
import {MatDialog} from "@angular/material";
import {
  ActivatedRoute,
  RouterStateSnapshot,
} from "@angular/router";
import {SharedAbstractDeleteDialog} from "@app/shared/components/dialogs/shared-abstract-delete/shared-abstract-delete.dialog";
import {SharedAbstractCrudRoutable} from "@app/shared/components/routables/shared-abstract-crud/shared-abstract-crud.routable";
import {DeleteDialog} from "@app/shared/models/delete-dialog.model";
import {StoreDialogUtil} from "@app/shared/utils/store-dialog.util";
import {StoreRouteLocalUtil} from "@app/shared/utils/store-route-local.util";
import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  Select,
  Store,
} from "@ngxs/store";
import {SharedAbstractFormPresentational} from "@shared/components/presentationals/shared-abstract-form/shared-abstract-form.presentational";
import {AppRoutesEnum} from "@shared/enums/routes.enum";
import {CrudHelper} from "@shared/helpers/crud.helper";
import {FormControlKey} from "@shared/models/form-control-key.model";
import {LocalStateModel} from "@shared/models/local-state.model";
import {Observable} from "rxjs";
import {
  distinctUntilChanged,
  map,
  take,
} from "rxjs/operators";
import {
  BaseResourceType,
  isNullOrUndefined,
  ModelFormControlEvent,
  QueryParameters,
  ResourceActionHelper,
  ResourceNameSpace,
  ResourceStateModel,
  StoreUtil,
} from "solidify-frontend";
import {LocalStateEnum} from "../../../enums/local-state.enum";

export abstract class SharedAbstractDetailEditRoutable<TResourceModel extends BaseResourceType, UResourceStateModel extends ResourceStateModel<TResourceModel>> extends SharedAbstractCrudRoutable<TResourceModel, UResourceStateModel> implements OnInit, OnDestroy {
  @Select((state: LocalStateModel) => state.router.state) urlStateObs: Observable<RouterStateSnapshot>;

  isLoadingObs: Observable<boolean> = this._store.select(s => StoreUtil.isLoadingState(super.getState(s)));
  currentObs: Observable<TResourceModel> = this._store.select(s => super.getState(s).current).pipe(
    tap(current => this.current = current),
  );
  current: TResourceModel;

  @ViewChild("formPresentational", {static: false})
  readonly formPresentational: SharedAbstractFormPresentational<any>;

  abstract readonly KEY_PARAM_NAME: keyof TResourceModel & string;

  protected _resId: string;

  isEdit: boolean;
  isEditObs: Observable<boolean>;

  protected constructor(protected readonly _store: Store,
                        protected readonly _route: ActivatedRoute,
                        protected readonly _actions$: Actions,
                        protected readonly _changeDetector: ChangeDetectorRef,
                        protected readonly _dialog: MatDialog,
                        protected readonly _state: LocalStateEnum,
                        protected readonly _resourceNameSpace: ResourceNameSpace,
                        protected readonly _parentState?: LocalStateEnum) {
    super(_store, _state, _parentState);
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.subscribe(this.getRetrieveEditModeObs());
  }

  ngOnDestroy(): void {
    super.ngOnDestroy();
  }

  protected retrieveCurrentModelAndResource(): void {
    this.retrieveResource();
    this.retrieveCurrentModelWithUrl();
  }

  protected retrieveResource(): void {
    this._store.dispatch(ResourceActionHelper.loadResource(this._resourceNameSpace));
  }

  protected cleanState(): void {
    this._store.dispatch(ResourceActionHelper.clean(this._resourceNameSpace, false, new QueryParameters()));
  }

  protected retrieveResIdFromUrl(): void {
    this._resId = this._route.snapshot.paramMap.get(AppRoutesEnum.paramIdWithoutPrefixParam);
    if (isNullOrUndefined(this._resId)) {
      this._resId = this._route.parent.snapshot.paramMap.get(AppRoutesEnum.paramIdWithoutPrefixParam);
    }
  }

  protected retrieveCurrentModelWithUrl(): void {
    this.retrieveResIdFromUrl();
    this.getById(this._resId);
  }

  protected getById(id: string): void {
    this._store.dispatch(ResourceActionHelper.getById(this._resourceNameSpace, id));
    this.getSubResourceWithParentId(id);
  }

  protected abstract getSubResourceWithParentId(id: string): void;

  protected getRetrieveEditModeObs(): Observable<boolean> {
    this.isEditObs = this.urlStateObs.pipe(
      map(url => url.url),
      distinctUntilChanged(),
      map(url => {
        this.isEdit = url.endsWith(AppRoutesEnum.separator + AppRoutesEnum.edit);
        this._changeDetector.detectChanges();
        return this.isEdit;
      }),
    );
    return this.isEditObs;
  }

  edit(): void {
    this._store.dispatch(new Navigate([StoreRouteLocalUtil.getEditRoute(this._state, this._resId)]));
  }

  update(model: ModelFormControlEvent<TResourceModel>): Observable<any> {
    super.saveInProgress();
    return this._store.dispatch(ResourceActionHelper.update<TResourceModel>(this._resourceNameSpace, model));
  }

  delete(): void {
    const datas = {} as DeleteDialog;
    this.subscribe(
      this.currentObs.pipe(take(1)),
      (model: TResourceModel) => {
        datas.name = model[this.KEY_PARAM_NAME];
        datas.resId = model.resId;
      });
    const dialogRef = this._dialog.open(StoreDialogUtil.delete(this._state) as ComponentType<SharedAbstractDeleteDialog<TResourceModel, UResourceStateModel>>, {
      width: "250px",
      data: datas,
    });
  }

  backToList(): void {
    this._store.dispatch(new Navigate([StoreRouteLocalUtil.getRootRoute(this._state)]));
  }

  backToDetail(): void {
    const currentUrl = this._store.selectSnapshot((s: LocalStateModel) => s.router.state.url);
    this._store.dispatch(new Navigate([StoreRouteLocalUtil.getDetailRoute(this._state), this._resId])).subscribe((state: LocalStateModel) => {
      if (state.router.state.url !== currentUrl) {
        // this.formPresentational.resetFormToInitialValue();
        // TODO : Fix don't need to get by id model from backend but fix resetFormToInitialValue with component user role org unit
        // TODO : Problem currently if redirect to detail page via breadcrumb in case of Deposit
        this.retrieveCurrentModelWithUrl();
      }
    });
  }

  checkAvailable(formControlKey: FormControlKey): void {
    this.subscribe(CrudHelper.checkAvailable(formControlKey, this._store, this._actions$, this._resourceNameSpace, this._resId));
  }
}
