import {ComponentType} from "@angular/cdk/overlay";
import {
  ChangeDetectorRef,
  OnInit,
} from "@angular/core";
import {MatDialog} from "@angular/material/dialog";
import {
  ActivatedRoute,
  UrlSegment,
} from "@angular/router";
import {SharedAbstractCrudRoutable} from "@app/shared/components/routables/shared-abstract-crud/shared-abstract-crud.routable";
import {DataTableColumns} from "@app/shared/models/data-table-columns.model";
import {StoreRouteLocalUtil} from "@app/shared/utils/store-route-local.util";
import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  ofActionCompleted,
  Store,
} from "@ngxs/store";
import {SharedAbstractDeleteDialog} from "@shared/components/dialogs/shared-abstract-delete/shared-abstract-delete.dialog";
import {LocalStateEnum} from "@shared/enums/local-state.enum";
import {RoutesEnum} from "@shared/enums/routes.enum";
import {DataTableActions} from "@shared/models/data-table-actions.model";
import {DeleteDialog} from "@shared/models/delete-dialog.model";
import {ExtraButtonToolbar} from "@shared/models/extra-button-toolbar.model";
import {StoreDialogUtil} from "@shared/utils/store-dialog.util";
import {Observable} from "rxjs";
import {tap} from "rxjs/operators";
import {
  BaseResourceType,
  QueryParameters,
  ResourceActionHelper,
  ResourceNameSpace,
  ResourceStateModel,
  StoreUtil,
  TRANSLATE,
} from "solidify-frontend";

export abstract class SharedAbstractListRoutable<TResourceModel extends BaseResourceType, UResourceStateModel extends ResourceStateModel<TResourceModel>> extends SharedAbstractCrudRoutable<TResourceModel, UResourceStateModel> implements OnInit {
  isLoadingObs: Observable<boolean>;
  listObs: Observable<TResourceModel[]>;
  queryParametersObs: Observable<QueryParameters>;

  skipInitialQuery: boolean = false;
  columnsSkippedToClear: string[] = [];

  columns: DataTableColumns<TResourceModel>[];
  actions: DataTableActions<TResourceModel>[];
  listNewId: string[];

  abstract readonly KEY_CREATE_BUTTON: string;
  abstract readonly KEY_REFRESH_BUTTON: string;
  abstract readonly KEY_BACK_BUTTON: string | undefined;
  abstract readonly KEY_PARAM_NAME: keyof TResourceModel & string;

  protected constructor(protected readonly _store: Store,
                        protected readonly _changeDetector: ChangeDetectorRef,
                        protected readonly _route: ActivatedRoute,
                        protected readonly _actions$: Actions,
                        protected readonly _dialog: MatDialog,
                        protected readonly _state: LocalStateEnum,
                        private readonly _resourceNameSpace: ResourceNameSpace,
                        public options?: SharedAbstractListRoutableOption<TResourceModel>,
                        protected readonly _parentState?: LocalStateEnum) {
    super(_store, _state, _parentState);
    this.options = Object.assign(this.getDefaultOptions(), options);
    this.isLoadingObs = this._store.select(s => StoreUtil.isLoadingState(super.getState(s)));
    this.listObs = this._store.select(s => this.getState(s).list);
    this.queryParametersObs = this._store.select(s => this.getState(s).queryParameters);
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.defineColumns();
    this.actions = this.defineActions();
  }

  private getDefaultOptions(): SharedAbstractListRoutableOption<TResourceModel> {
    return {
      canCreate: true,
      canGoBack: true,
      canRefresh: true,
    };
  }

  abstract defineColumns(): void;

  protected defineActions(): DataTableActions<TResourceModel>[] {
    return [
      {
        logo: "edit",
        callback: (model: TResourceModel) => this.goToEdit(model),
        placeholder: TRANSLATE("crud.list.action.goToEdit"),
        displayOnCondition: (model: TResourceModel) => this.conditionDisplayEditButton(model),
      },
      {
        logo: "delete",
        callback: (model: TResourceModel) => this.delete(model),
        placeholder: TRANSLATE("crud.list.action.delete"),
        displayOnCondition: (model: TResourceModel) => this.conditionDisplayDeleteButton(model),
      },
    ];
  }

  abstract conditionDisplayEditButton(model: TResourceModel | undefined): boolean;

  abstract conditionDisplayDeleteButton(model: TResourceModel | undefined): boolean;

  goToEdit(model: TResourceModel): void {
    this._store.dispatch(new Navigate([StoreRouteLocalUtil.getEditRoute(this._state, model.resId)]));
  }

  delete(model: TResourceModel): void {
    const datas = {} as DeleteDialog;
    datas.name = model[this.KEY_PARAM_NAME];
    datas.resId = model.resId;

    const dialogRef = this._dialog.open(StoreDialogUtil.delete(this._state) as ComponentType<SharedAbstractDeleteDialog<TResourceModel, UResourceStateModel>>, {
      width: "250px",
      data: datas,
    });
    this.subscribe(this._actions$.pipe(ofActionCompleted(this._resourceNameSpace.DeleteSuccess))
      .pipe(
        tap(result => {
          if (result.result.successful) {
            this.getAll();
          }
        }),
      ));
  }

  create(): void {
    this.navigate(StoreRouteLocalUtil.getCreateRoute(this._state));
  }

  getAll(queryParameters?: QueryParameters): void {
    this._store.dispatch(ResourceActionHelper.getAll(this._resourceNameSpace, queryParameters, true));
  }

  showDetail(model: TResourceModel): void {
    this._store.dispatch(new Navigate([StoreRouteLocalUtil.getDetailRoute(this._state), model.resId]));
  }

  onQueryParametersEvent(queryParameters: QueryParameters): void {
    this._store.dispatch(ResourceActionHelper.changeQueryParameters(this._resourceNameSpace, queryParameters, true));
    this._changeDetector.detectChanges(); // Allow to display spinner the first time
  }

  back(): void {
    const urlSegmentToParent: UrlSegment[] = [];
    this._route.snapshot.parent.parent.pathFromRoot.forEach((activatedRouteSnap) => {
      urlSegmentToParent.push(...activatedRouteSnap.url);
    });
    this._store.dispatch(new Navigate(urlSegmentToParent.map(u => u.path)));
  }

  navigate(route: RoutesEnum): void {
    this._store.dispatch(new Navigate([route]));
  }
}

export interface SharedAbstractListRoutableOption<TResourceModel> {
  canCreate?: boolean;
  canGoBack?: boolean;
  canRefresh?: boolean;
  listExtraButtons?: ExtraButtonToolbar<TResourceModel>[];
}
